

#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <math.h>
#include "matrix_lib.h"

#define N 80

void fill_matrix(double **A, int start_m, int stop_m, int start_n, int stop_n);
void fill_vector(double *v, int start, int stop);

/**
 * the program creates the NxN-matrix and defines it by 
 * default to be 1 at every entry; the matrix is multiplied
 * with the vector v=[1, 1, ..., 1] using 2D partitioning  
 * which gives the expected output vector x=[n, n, ..., n];
 * this is printed for each of the diagonal elements
 */

int main(int argc, char *argv[])
{
  double **matrixA, *v, *u, *x; 
  int i, j, n;

  int my_rank, my_cart_rank, num_procs, np;
  int dims[2], periods[2], my_coords[2];
  int start_n, my_n, start_m, my_m;
  
  MPI_Comm mpi_comm_cart;
  MPI_Comm mpi_comm_cols;
  MPI_Comm mpi_comm_rows;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &num_procs);
  
  /* check if num_procs is square number */
  if((int)sqrt(num_procs)*(int)sqrt(num_procs)!=num_procs) { 
    if(my_rank==0)
      fprintf(stderr, "Please give a square number of processors.");
    MPI_Abort(MPI_COMM_WORLD, 1);
  }

  if(num_procs>N*N) {
    if(my_rank==0)
      fprintf(stderr, "Please give number of processors less than sqrt(N)\n");
    MPI_Abort(MPI_COMM_WORLD, 1);
  }

  np = (int)sqrt(num_procs);             // have np*np processors
  n = N;
  dims[0] = dims[1] = np;                // p*p partitioning
  periods[0] = periods[1] = 1;           // keep structure

  /* create cartesian structure */
  MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, 1, &mpi_comm_cart);
  MPI_Comm_rank(mpi_comm_cart, &my_cart_rank);
  MPI_Cart_coords(mpi_comm_cart, my_cart_rank, 2, my_coords);
  
  /* keep structure in each direction */
  MPI_Cart_sub(mpi_comm_cart, (int[]){0, 1}, &mpi_comm_rows);
  MPI_Cart_sub(mpi_comm_cart, (int[]){1, 0}, &mpi_comm_cols);

  /* calculate coordinates */
  my_m = n/np + (n%np>my_coords[0]);   // m = n since A is square
  start_m = 0;
  for(i = 0; i < my_coords[0]; i++)
    start_m+=n/np + (n%np>i);

  my_n = n/np + (n%np>my_coords[1]);
  start_n = 0;
  for(i = 0; i < my_coords[1]; i++)
    start_n+=n/np + (n%np>i);
  
  allocate_matrix(&matrixA, my_m, my_n);
  allocate_vector(&v, my_n);
  allocate_vector(&u, my_m);
  allocate_vector(&x, my_m);

  /* each process calculates it's own part of A and the diagonal squares the vector v */
  fill_matrix(matrixA, start_m, start_m+my_m, start_n, start_n+my_n);
  if(my_coords[0]==my_coords[1])
    fill_vector(v, start_n, start_n+my_n);

  /* broadcast vector columnwise */
  MPI_Bcast(v, my_n, MPI_DOUBLE, my_coords[1], mpi_comm_cols);
  
  /* calculate own sum */
  for(i = 0; i < my_m; i++) {
    u[i] = 0;
    for(j = 0; j < my_n; j++) {
      u[i]+=matrixA[i][j]*v[i];
    }
  }

  /* reduce rowwise to diagonal procs */
  MPI_Reduce(u, x, my_m, MPI_DOUBLE, MPI_SUM, my_coords[0], mpi_comm_rows);
  
  if(my_coords[0]==my_coords[1]) {
    for(i = 0; i < my_n; i++) {
      printf("%f \n", x[i]);
    }
  }

  MPI_Comm_free(&mpi_comm_cart);
  MPI_Comm_free(&mpi_comm_rows);
  MPI_Comm_free(&mpi_comm_cols);

  deallocate_matrix(matrixA);
  deallocate_vector(v);
  deallocate_vector(u);
  deallocate_vector(x);

  MPI_Finalize();
  return 0;
}

void fill_matrix(double **A, int start_m, int stop_m, int start_n, int stop_n) {
  int i, j;
  for(i = 0; i < stop_m-start_m; i++) {
    for(j = 0; j < stop_n-start_n; j++) {
      A[i][j] = 1;
    }
  }
}

void fill_vector(double *v, int start, int stop) {
  int i;
  for(i = 0; i < stop-start; i++)
    v[i] = 1;
}


