

#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <math.h>
#include <string.h>
#include "matrix_lib.h"

#define N 200

void fill_matrix_A(double **A, int start_m, int stop_m, int start_n, int stop_n, int ub);
void fill_matrix_B(double **B, int start_m, int stop_m, int start_n, int stop_n, int ub);
void multiply_matrices(double **matrixA, double **matrixB,
		       double **matrixC, int m, int l, int n);

/**
 * the program creates the NxN-matrices A and B where B
 * is set to be the I matrix and A the matrix
 * [0, 1, ...(n-1); n .... n+n-1; ... ; ... (n-1)*(n-1)]
 * these two are multiplied using 2D partitioning with
 * imaginary rows/columns if number of processes is not a
 * multiple of N (to simplify communication)
 * result should be A*I = A
 */

int main(int argc, char *argv[])
{
  double **matrixA, **matrixB, **matrixC;
  double ***cubeA, ***cubeB;
  int k, counter, n, upper_bound;
  int *disp, *recv_count;

  int my_rank, my_cart_rank, num_procs, np;
  int dims[2], periods[2], my_coords[2];
  int start_n, my_n, start_m, my_m;
  
  MPI_Comm mpi_comm_cart;
  MPI_Comm mpi_comm_cols;
  MPI_Comm mpi_comm_rows;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &num_procs);
  
  /* check if num_procs is square number */
  if((int)sqrt(num_procs)*(int)sqrt(num_procs)!=num_procs) { 
    if(my_rank==0)
      fprintf(stderr, "Please give a square number of processors.");
    MPI_Abort(MPI_COMM_WORLD, 1);
  }

  if(num_procs>N*N) {
    if(my_rank==0)
      fprintf(stderr, "Please give number of processors less than sqrt(N)\n");
    MPI_Abort(MPI_COMM_WORLD, 1);
  }

  np = (int)sqrt(num_procs);             // have npxnp processors
  n = N;                                 // and nxn matrices
  dims[0] = dims[1] = np;              
  periods[0] = periods[1] = 1;           // keep structure

  /* create cartesian structure */
  MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, 1, &mpi_comm_cart);
  MPI_Comm_rank(mpi_comm_cart, &my_cart_rank);
  MPI_Cart_coords(mpi_comm_cart, my_cart_rank, 2, my_coords);
  
  /* keep structure in each direction */
  MPI_Cart_sub(mpi_comm_cart, (int[]){0, 1}, &mpi_comm_rows);
  MPI_Cart_sub(mpi_comm_cart, (int[]){1, 0}, &mpi_comm_cols);

  /* calculate coordinates; m = n since A and B are square 
     so that everyone know all dimensions of real partition */
  disp = malloc(np*sizeof(int));
  recv_count = malloc(np*sizeof(int));

  counter = 0;
  for(k = 0; k < np; k++) {
    recv_count[k] = n*((n/np) + (n%np>k));
    disp[k] = counter;
    counter+=recv_count[k];    
  }
  
  /* variables used to simplify readibility */
  upper_bound = recv_count[0]/n;         
  my_m = recv_count[my_coords[0]]/n;
  my_n = recv_count[my_coords[1]]/n;
  start_m = disp[my_coords[0]]/n;
  start_n = disp[my_coords[1]]/n;

  /* allocating space for each submatrix */
  allocate_matrix(&matrixA, my_m, upper_bound);
  allocate_matrix(&matrixB, upper_bound, my_n);
  allocate_matrix(&matrixC, my_m, my_n);

  allocate_cube(&cubeA, np, my_m, upper_bound);
  allocate_cube(&cubeB, np, upper_bound, my_n);

  /* each process calculates it's own part of A and B before broadcasting */
  fill_matrix_A(matrixA, start_m, start_m+my_m,
		start_n, start_n+my_n, upper_bound);
  fill_matrix_B(matrixB, start_m, start_m+my_m, 
		start_n, start_n+my_n, upper_bound);
       
  MPI_Allgather(matrixA[0], my_m*upper_bound, MPI_DOUBLE, cubeA[0][0],
		my_m*upper_bound, MPI_DOUBLE, mpi_comm_rows);
  
  MPI_Allgather(matrixB[0], my_n*upper_bound, MPI_DOUBLE, cubeB[0][0],
		my_n*upper_bound, MPI_DOUBLE, mpi_comm_cols);

  /* multiplication*/  
  memset(matrixC[0], 0, my_m*my_n*sizeof(double));
  
  for(k = 0; k < np; k++)
    multiply_matrices(cubeA[k], cubeB[k], matrixC,
		      my_m, recv_count[k]/n, my_n);

  MPI_Comm_free(&mpi_comm_cart);
  MPI_Comm_free(&mpi_comm_rows);
  MPI_Comm_free(&mpi_comm_cols);

  deallocate_matrix(matrixA);
  deallocate_matrix(matrixB);
  deallocate_matrix(matrixC);
  deallocate_cube(cubeA);
  deallocate_cube(cubeB);

  MPI_Finalize();
  return 0;
}

/** matrix 1, 2, 3 ... (n-1)(n-1) */
void fill_matrix_A(double **A, int start_m, int stop_m,
		   int start_n, int stop_n, int ub) {
  int i, j;
  for(i = 0; i < stop_m-start_m; i++) {
    for(j = 0; j < stop_n-start_n; j++) {
      A[i][j] = (i+start_m)*N+j+start_n;
    }
    for(j = stop_n-start_n; j < ub; j++) {
      A[i][j] = 0;   // imaginary column
    }
  }
}

/** I matrix */
void fill_matrix_B(double **B, int start_m, int stop_m,
		   int start_n, int stop_n, int ub) {
  int i, j;
  for(i = 0; i < stop_m-start_m; i++) {
    for(j = 0; j < stop_n-start_n; j++) {
      if(i+start_m==j+start_n)
	B[i][j] = 1;
      else
	B[i][j] = 0;
    }
  }
  for(i = stop_m-start_m; i < ub; i++) {
    B[i][stop_n-start_n-1] = 0;  // imaginary row
  }
}


/** multiply the matrices A and B
 * A and B must be of dimensions m*l, l*n, C of dimension m*n,
 * C must have initialized values
 * calculation can be performed several times on same result matrix
 */
void multiply_matrices(double **matrixA, double **matrixB, double **matrixC, int m, int l, int n)
{
  int i, j, k;
  for(i = 0; i < m; i++) {
    for(j = 0; j < n; j++) {
      for(k = 0; k < l; k++) {  
	matrixC[i][j] += matrixA[i][k]*matrixB[k][j];
      }
    }
  } 
}
